import os
from typing import List

from pydantic import BaseModel

from swarms.models.openai_function_caller import OpenAIFunctionCaller
from swarms import OpenAIChat
from swarms.structs.agent import Agent
from swarms.structs.concat import concat_strings

api_key = os.getenv("OPENAI_API_KEY")

# Create an instance of the OpenAIChat class
model = OpenAIChat(
    api_key=api_key, model_name="gpt-4o-mini", temperature=0.1
)


# Initialize the agents
growth_agent1 = Agent(
    agent_name="marketing_specialist",
    system_prompt="You're the marketing specialist, your purpose is to help companies grow by improving their marketing strategies!",
    agent_description="Improve a company's marketing strategies!",
    llm=model,
    max_loops=1,
    autosave=True,
    dashboard=False,
    verbose=True,
    streaming_on=True,
    saved_state_path="marketing_specialist.json",
    stopping_token="Stop!",
    context_length=1000,
)

growth_agent2 = Agent(
    agent_name="sales_specialist",
    system_prompt="You're the sales specialist, your purpose is to help companies grow by improving their sales strategies!",
    agent_description="Improve a company's sales strategies!",
    llm=model,
    max_loops=1,
    autosave=True,
    dashboard=False,
    verbose=True,
    streaming_on=True,
    saved_state_path="sales_specialist.json",
    stopping_token="Stop!",
    context_length=1000,
)

growth_agent3 = Agent(
    agent_name="product_development_specialist",
    system_prompt="You're the product development specialist, your purpose is to help companies grow by improving their product development strategies!",
    agent_description="Improve a company's product development strategies!",
    llm=model,
    max_loops=1,
    autosave=True,
    dashboard=False,
    verbose=True,
    streaming_on=True,
    saved_state_path="product_development_specialist.json",
    stopping_token="Stop!",
    context_length=1000,
)

team = [growth_agent1, growth_agent2, growth_agent3]


# class HiearchicalSwarm(BaseModel):
#     agents: List[Agent]
#     director: Agent
#     planner: Agent
#     max_loops: int = 3
#     verbose: bool = True

#     def run(self, task: str):
#         # Plan
#         # Plan -> JSON Function call -> workers -> response fetch back to boss -> planner
#         responses = []
#         responses.append(task)

#         for _ in range(self.max_loops):
#             # Plan
#             plan = self.planner.run(concat_strings(responses))
#             logger.info(f"Agent {self.planner.agent_name} planned: {plan}")
#             responses.append(plan)

#             # Execute json function calls
#             calls = self.director.run(plan)
#             logger.info(
#                 f"Agent {self.director.agent_name} called: {calls}"
#             )
#             responses.append(calls)
#             # Parse and send tasks to agents
#             output = parse_then_send_tasks_to_agents(self.agents, calls)

#             # Fetch back to boss
#             responses.append(output)

#         return concat_strings(responses)

#     def __call__(self, task: str):
#         responses = []
#         responses.append(task)

#         for _ in range(self.max_loops):
#             output = self.step(task, responses)
#             responses.append(output)

#         return concat_strings(responses)

#     def step(self, responses: List[str] = None) -> str:
#         # Plan
#         # Plan -> JSON Function call -> workers -> response fetch back to boss -> planner

#         # Plan
#         plan = self.planner.run(concat_strings(responses))
#         logger.info(f"Agent {self.planner.agent_name} planned: {plan}")
#         responses.append(plan)

#         # Execute json function calls
#         calls = self.director.run(plan)
#         logger.info(f"Agent {self.director.agent_name} called: {calls}")
#         responses.append(calls)
#         # Parse and send tasks to agents
#         output = parse_then_send_tasks_to_agents(self.agents, calls)

#         # Fetch back to boss
#         responses.append(output)

#         return concat_strings(responses)

#     def plan(self, task: str, responses: List[str] = None):
#         # Plan
#         # Plan -> JSON Function call -> workers -> response fetch back to boss -> planner
#         # responses = []
#         # responses.append(task)

#         # Plan
#         plan = self.planner.run(concat_strings(responses))
#         logger.info(f"Agent {self.planner.agent_name} planned: {plan}")
#         responses.append(plan)

#         return concat_strings(responses)


def agents_list(
    agents: List[Agent] = team,
) -> str:
    responses = []

    for agent in agents:
        name = agent.agent_name
        description = agent.description
        response = f"Agent Name {name}: Description {description}"
        responses.append(response)

    return concat_strings(responses)


def parse_then_send_tasks_to_agents(agents: List[Agent], response: dict):
    # Initialize an empty dictionary to store the output of each agent
    output = []

    # Loop over the tasks in the response
    for call in response["calls"]:
        name = call["agent_name"]
        task = call["task"]

        # Loop over the agents
        for agent in agents:
            # If the agent's name matches the name in the task, run the task
            if agent.agent_name == name:
                out = agent.run(task)
                print(out)

                output.append(f"{name}: {out}")

                # Store the output in the dictionary
                # output[name] = out
                break

    return output


class HierarchicalOrderCall(BaseModel):
    agent_name: str
    task: str


class CallTeam(BaseModel):
    calls: List[HierarchicalOrderCall]


# Example usage:
system_prompt = f"""
You're a director agent, your responsibility is to serve the user efficiently, effectively and skillfully.You have a swarm of agents available to distribute tasks to, interact with the user and then submit tasks to the worker agents. Provide orders to the worker agents that are direct, explicit, and simple. Ensure that they are given tasks that are understandable, actionable, and simple to execute.


######
Workers available:

{agents_list(team)}


"""


# Initialize the function caller
function_caller = OpenAIFunctionCaller(
    system_prompt=system_prompt,
    openai_api_key=os.getenv("OPENAI_API_KEY"),
    max_tokens=500,
    temperature=0.5,
    base_model=CallTeam,
)

# Run the function caller
response = function_caller.run(
    "Now let's grow the company! Send an order to the marketing specialist, sales specialist, and product development specialist to improve the company's growth strategies."
)
# print(response)
print(response)
print(type(response))


out = parse_then_send_tasks_to_agents(team, response)
print(out)



import requests
from loguru import logger
import os


def fetch_secrets_from_vault(
    client_id: str = os.getenv("HCP_CLIENT_ID"),
    client_secret: str = os.getenv("HCP_CLIENT_SECRET"),
    organization_id: str = os.getenv("HCP_ORGANIZATION_ID"),
    project_id: str = os.getenv("HCP_PROJECT_ID"),
    app_id: str = os.getenv("HCP_APP_ID"),
) -> str:
    """
    Fetch secrets from HashiCorp Vault using service principal authentication.

    Args:
        client_id (str): The client ID for the service principal.
        client_secret (str): The client secret for the service principal.
        organization_id (str): The ID of the organization in HCP.
        project_id (str): The ID of the project in HCP.
        app_id (str): The ID of the app in HCP.

    Returns:
        str: A dictionary containing the fetched secrets.

    Raises:
        Exception: If there is an error retrieving the API token or secrets.
    """
    # Step 1: Generate the API Token
    token_url = "https://auth.idp.hashicorp.com/oauth2/token"
    token_data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "grant_type": "client_credentials",
        "audience": "https://api.hashicorp.cloud",
    }
    token_headers = {"Content-Type": "application/x-www-form-urlencoded"}

    logger.info("Requesting API token from HashiCorp Vault")
    response = requests.post(
        token_url, data=token_data, headers=token_headers
    )

    if response.status_code != 200:
        logger.error(
            f"Failed to retrieve API token. Status Code: {response.status_code}, Response: {response.text}"
        )
        response.raise_for_status()

    api_token = response.json().get("access_token")

    if not api_token:
        raise Exception("Failed to retrieve API token")

    # Step 2: Fetch Secrets
    secrets_url = f"https://api.cloud.hashicorp.com/secrets/2023-06-13/organizations/{organization_id}/projects/{project_id}/apps/{app_id}/open"
    secrets_headers = {"Authorization": f"Bearer {api_token}"}

    logger.info("Fetching secrets from HashiCorp Vault")
    response = requests.get(secrets_url, headers=secrets_headers)

    if response.status_code != 200:
        logger.error(
            f"Failed to fetch secrets. Status Code: {response.status_code}, Response: {response.text}"
        )
        response.raise_for_status()

    secrets = response.json()

    for secret in secrets["secrets"]:
        name = secret.get("name")
        value = secret.get("version", {}).get("value")
        print(f"Name: {name}, Value: {value}")

    return name, value


# def main() -> None:
#     """
#     Main function to fetch secrets from HashiCorp Vault and print them.

#     Raises:
#         EnvironmentError: If required environment variables are not set.
#     """
#     HCP_CLIENT_ID = os.getenv("HCP_CLIENT_ID")
#     HCP_CLIENT_SECRET = os.getenv("HCP_CLIENT_SECRET")
#     ORGANIZATION_ID = os.getenv("HCP_ORGANIZATION_ID")
#     PROJECT_ID = os.getenv("HCP_PROJECT_ID")
#     APP_ID = os.getenv("HCP_APP_ID")

#     # if not all([HCP_CLIENT_ID, HCP_CLIENT_SECRET, ORGANIZATION_ID, PROJECT_ID, APP_ID]):
#     #     raise EnvironmentError("One or more environment variables are missing: HCP_CLIENT_ID, HCP_CLIENT_SECRET, ORGANIZATION_ID, PROJECT_ID, APP_ID")

#     secrets = fetch_secrets_from_vault(
#         HCP_CLIENT_ID,
#         HCP_CLIENT_SECRET,
#         ORGANIZATION_ID,
#         PROJECT_ID,
#         APP_ID,
#     )
#     print(secrets)

# for secret in secrets["secrets"]:
#     name = secret.get("name")
#     value = secret.get("version", {}).get("value")
#     print(f"Name: {name}, Value: {value}")


# if __name__ == "__main__":
#     main()



from __future__ import annotations

import asyncio
import json
import os
import platform
from typing import Any

import pkg_resources
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import Span, Status, StatusCode
from swarms.structs.base_swarm import BaseSwarm


class Telemetry:
    """A class to handle anonymous telemetry for the swarms package.

    The data being collected is for development purpose, all data is anonymous.

    There is NO data being collected on the prompts, tasks descriptions
    agents backstories or goals nor responses or any data that is being
    processed by the agents, nor any secrets and env vars.

    Data collected includes:
    - Version of swarms
    - Version of Python
    - General OS (e.g. number of CPUs, macOS/Windows/Linux)
    - Number of agents and tasks in a crew
    - Crew Process being used
    - If Agents are using memory or allowing delegation
    - If Tasks are being executed in parallel or sequentially
    - Language model being used
    - Roles of agents in a crew
    - Tools names available

    Users can opt-in to sharing more complete data using the `share_crew`
    attribute in the Crew class.
    """

    def __init__(self):
        self.ready = False
        self.trace_set = False
        try:
            telemetry_endpoint = "https://telemetry.swarms.com:4319"
            self.resource = Resource(
                attributes={SERVICE_NAME: "swarms-telemetry"},
            )
            self.provider = TracerProvider(resource=self.resource)

            processor = BatchSpanProcessor(
                OTLPSpanExporter(
                    endpoint=f"{telemetry_endpoint}/v1/traces",
                    timeout=30,
                )
            )

            self.provider.add_span_processor(processor)
            self.ready = True
        except BaseException as e:
            if isinstance(
                e,
                (
                    SystemExit,
                    KeyboardInterrupt,
                    GeneratorExit,
                    asyncio.CancelledError,
                ),
            ):
                raise  # Re-raise the exception to not interfere with system signals
            self.ready = False

    def set_tracer(self):
        if self.ready and not self.trace_set:
            try:
                trace.set_tracer_provider(self.provider)
                self.trace_set = True
            except Exception:
                self.ready = False
                self.trace_set = False

    def swarm_creation(
        self, swarm: BaseSwarm, inputs: dict[str, Any] | None
    ):
        """Records the creation of a crew."""
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Crew Created")
                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(
                    span, "python_version", platform.python_version()
                )
                self._add_attribute(span, "crew_key", crew.key)
                self._add_attribute(span, "crew_id", str(crew.id))
                self._add_attribute(span, "crew_process", crew.process)
                self._add_attribute(span, "crew_memory", crew.memory)
                self._add_attribute(
                    span, "crew_number_of_tasks", len(crew.tasks)
                )
                self._add_attribute(
                    span, "crew_number_of_agents", len(crew.agents)
                )
                self._add_attribute(
                    span,
                    "crew_agents",
                    json.dumps(
                        [
                            {
                                "key": agent.key,
                                "id": str(agent.id),
                                "role": agent.role,
                                "goal": agent.goal,
                                "backstory": agent.backstory,
                                "verbose?": agent.verbose,
                                "max_iter": agent.max_iter,
                                "max_rpm": agent.max_rpm,
                                "i18n": agent.i18n.prompt_file,
                                "llm": json.dumps(
                                    self._safe_llm_attributes(agent.llm)
                                ),
                                "delegation_enabled?": agent.allow_delegation,
                                "tools_names": [
                                    tool.name.casefold()
                                    for tool in agent.tools or []
                                ],
                            }
                            for agent in crew.agents
                        ]
                    ),
                )
                self._add_attribute(
                    span,
                    "crew_tasks",
                    json.dumps(
                        [
                            {
                                "key": task.key,
                                "id": str(task.id),
                                "description": task.description,
                                "expected_output": task.expected_output,
                                "async_execution?": task.async_execution,
                                "human_input?": task.human_input,
                                "agent_role": (
                                    task.agent.role
                                    if task.agent
                                    else "None"
                                ),
                                "agent_key": (
                                    task.agent.key if task.agent else None
                                ),
                                "context": (
                                    [
                                        task.description
                                        for task in task.context
                                    ]
                                    if task.context
                                    else None
                                ),
                                "tools_names": [
                                    tool.name.casefold()
                                    for tool in task.tools or []
                                ],
                            }
                            for task in crew.tasks
                        ]
                    ),
                )
                self._add_attribute(span, "platform", platform.platform())
                self._add_attribute(
                    span, "platform_release", platform.release()
                )
                self._add_attribute(
                    span, "platform_system", platform.system()
                )
                self._add_attribute(
                    span, "platform_version", platform.version()
                )
                self._add_attribute(span, "cpus", os.cpu_count())

                if crew.share_crew:
                    self._add_attribute(
                        span,
                        "crew_inputs",
                        json.dumps(inputs) if inputs else None,
                    )

                span.set_status(Status(StatusCode.OK))
                span.end()
            except Exception:
                pass

    def task_started(self, crew: Crew, task: Task) -> Span | None:
        """Records task started in a crew."""
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")

                created_span = tracer.start_span("Task Created")

                self._add_attribute(created_span, "crew_key", crew.key)
                self._add_attribute(created_span, "crew_id", str(crew.id))
                self._add_attribute(created_span, "task_key", task.key)
                self._add_attribute(created_span, "task_id", str(task.id))

                if crew.share_crew:
                    self._add_attribute(
                        created_span,
                        "formatted_description",
                        task.description,
                    )
                    self._add_attribute(
                        created_span,
                        "formatted_expected_output",
                        task.expected_output,
                    )

                created_span.set_status(Status(StatusCode.OK))
                created_span.end()

                span = tracer.start_span("Task Execution")

                self._add_attribute(span, "crew_key", crew.key)
                self._add_attribute(span, "crew_id", str(crew.id))
                self._add_attribute(span, "task_key", task.key)
                self._add_attribute(span, "task_id", str(task.id))

                if crew.share_crew:
                    self._add_attribute(
                        span, "formatted_description", task.description
                    )
                    self._add_attribute(
                        span,
                        "formatted_expected_output",
                        task.expected_output,
                    )

                return span
            except Exception:
                pass

        return None

    def task_ended(self, span: Span, task: Task, crew: Crew):
        """Records task execution in a crew."""
        if self.ready:
            try:
                if crew.share_crew:
                    self._add_attribute(
                        span,
                        "task_output",
                        task.output.raw if task.output else "",
                    )

                span.set_status(Status(StatusCode.OK))
                span.end()
            except Exception:
                pass

    def tool_repeated_usage(self, llm: Any, tool_name: str, attempts: int):
        """Records the repeated usage 'error' of a tool by an agent."""
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Tool Repeated Usage")
                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(span, "tool_name", tool_name)
                self._add_attribute(span, "attempts", attempts)
                if llm:
                    self._add_attribute(
                        span,
                        "llm",
                        json.dumps(self._safe_llm_attributes(llm)),
                    )
                span.set_status(Status(StatusCode.OK))
                span.end()
            except Exception:
                pass

    def tool_usage(self, llm: Any, tool_name: str, attempts: int):
        """Records the usage of a tool by an agent."""
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Tool Usage")
                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(span, "tool_name", tool_name)
                self._add_attribute(span, "attempts", attempts)
                if llm:
                    self._add_attribute(
                        span,
                        "llm",
                        json.dumps(self._safe_llm_attributes(llm)),
                    )
                span.set_status(Status(StatusCode.OK))
                span.end()
            except Exception:
                pass

    def tool_usage_error(self, llm: Any):
        """Records the usage of a tool by an agent."""
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Tool Usage Error")
                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                if llm:
                    self._add_attribute(
                        span,
                        "llm",
                        json.dumps(self._safe_llm_attributes(llm)),
                    )
                span.set_status(Status(StatusCode.OK))
                span.end()
            except Exception:
                pass

    def individual_test_result_span(
        self, crew: Crew, quality: int, exec_time: int, model_name: str
    ):
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Crew Individual Test Result")

                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(span, "crew_key", crew.key)
                self._add_attribute(span, "crew_id", str(crew.id))
                self._add_attribute(span, "quality", str(quality))
                self._add_attribute(span, "exec_time", str(exec_time))
                self._add_attribute(span, "model_name", model_name)
                return span
            except Exception:
                pass

    def test_execution_span(
        self,
        crew: Crew,
        iterations: int,
        inputs: dict[str, Any] | None,
        model_name: str,
    ):
        if self.ready:
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Crew Test Execution")

                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(span, "crew_key", crew.key)
                self._add_attribute(span, "crew_id", str(crew.id))
                self._add_attribute(span, "iterations", str(iterations))
                self._add_attribute(span, "model_name", model_name)

                if crew.share_crew:
                    self._add_attribute(
                        span,
                        "inputs",
                        json.dumps(inputs) if inputs else None,
                    )

                return span
            except Exception:
                pass

    def crew_execution_span(
        self, crew: Crew, inputs: dict[str, Any] | None
    ):
        """Records the complete execution of a crew.
        This is only collected if the user has opted-in to share the crew.
        """
        self.crew_creation(crew, inputs)

        if (self.ready) and (crew.share_crew):
            try:
                tracer = trace.get_tracer("swarms.telemetry")
                span = tracer.start_span("Crew Execution")
                self._add_attribute(
                    span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(span, "crew_key", crew.key)
                self._add_attribute(span, "crew_id", str(crew.id))
                self._add_attribute(
                    span,
                    "crew_inputs",
                    json.dumps(inputs) if inputs else None,
                )
                self._add_attribute(
                    span,
                    "crew_agents",
                    json.dumps(
                        [
                            {
                                "key": agent.key,
                                "id": str(agent.id),
                                "role": agent.role,
                                "goal": agent.goal,
                                "backstory": agent.backstory,
                                "verbose?": agent.verbose,
                                "max_iter": agent.max_iter,
                                "max_rpm": agent.max_rpm,
                                "i18n": agent.i18n.prompt_file,
                                "llm": json.dumps(
                                    self._safe_llm_attributes(agent.llm)
                                ),
                                "delegation_enabled?": agent.allow_delegation,
                                "tools_names": [
                                    tool.name.casefold()
                                    for tool in agent.tools or []
                                ],
                            }
                            for agent in crew.agents
                        ]
                    ),
                )
                self._add_attribute(
                    span,
                    "crew_tasks",
                    json.dumps(
                        [
                            {
                                "id": str(task.id),
                                "description": task.description,
                                "expected_output": task.expected_output,
                                "async_execution?": task.async_execution,
                                "human_input?": task.human_input,
                                "agent_role": (
                                    task.agent.role
                                    if task.agent
                                    else "None"
                                ),
                                "agent_key": (
                                    task.agent.key if task.agent else None
                                ),
                                "context": (
                                    [
                                        task.description
                                        for task in task.context
                                    ]
                                    if task.context
                                    else None
                                ),
                                "tools_names": [
                                    tool.name.casefold()
                                    for tool in task.tools or []
                                ],
                            }
                            for task in crew.tasks
                        ]
                    ),
                )
                return span
            except Exception:
                pass

    def end_crew(self, crew, final_string_output):
        if (self.ready) and (crew.share_crew):
            try:
                self._add_attribute(
                    crew._execution_span,
                    "swarms_version",
                    pkg_resources.get_distribution("swarms").version,
                )
                self._add_attribute(
                    crew._execution_span,
                    "crew_output",
                    final_string_output,
                )
                self._add_attribute(
                    crew._execution_span,
                    "crew_tasks_output",
                    json.dumps(
                        [
                            {
                                "id": str(task.id),
                                "description": task.description,
                                "output": task.output.raw_output,
                            }
                            for task in crew.tasks
                        ]
                    ),
                )
                crew._execution_span.set_status(Status(StatusCode.OK))
                crew._execution_span.end()
            except Exception:
                pass

    def _add_attribute(self, span, key, value):
        """Add an attribute to a span."""
        try:
            return span.set_attribute(key, value)
        except Exception:
            pass

    def _safe_llm_attributes(self, llm):
        attributes = [
            "name",
            "model_name",
            "base_url",
            "model",
            "top_k",
            "temperature",
        ]
        if llm:
            safe_attributes = {
                k: v for k, v in vars(llm).items() if k in attributes
            }
            safe_attributes["class"] = llm.__class__.__name__
            return safe_attributes
        return {}
